from PIL import ImageFont, ImageDraw, Image
from skimage.metrics import structural_similarity as ssim
import numpy as np
import logging
import sys
import os

from . import font


FONT_SIZE = 20

font_path = font.get_monospace_font()
font_cache = {}
char_cache = {}

def get_font(font_path, size):
    key = (font_path, size)
    if key not in font_cache:
        font_cache[key] = ImageFont.truetype(font_path, size)
    return font_cache[key]

# from UCYT5040/lectrick
def crop_image(image):
    image_array = np.array(image)
    non_white_mask = image_array < 255
    non_white_pixels = np.argwhere(non_white_mask)
    if non_white_pixels.size == 0:
        return image
    top, left = non_white_pixels.min(axis=0)
    bottom, right = non_white_pixels.max(axis=0) + 1  # +1 to include the last pixel
    return image.crop((left, top, right, bottom))

def get_char(char, width, height, save=False):
    char_arr = char_cache.get(char)
    if char_arr is not None:
        return char_arr
    else:
        font = get_font(font_path, FONT_SIZE)
        bbox = font.getbbox(char)
        
        if not (bbox and bbox[2] > bbox[0] and bbox[3] > bbox[1]):
            return None
        
        char_width, char_height = bbox[2] - bbox[0], bbox[3] - bbox[1]
        
        char_image = Image.new('L', (char_width, char_height), 255)
        draw = ImageDraw.Draw(char_image)
        draw.text((-bbox[0], -bbox[1]), char, font=font, fill=0)
                
        char_image = crop_image(char_image)
        
        char_image = char_image.resize((width, height), Image.Resampling.LANCZOS)
        
        if save:
            os.makedirs("chars", exist_ok=True)
            char_code = ord(char)
            char_image.save(f"chars/{char_code}.png")
        
        char_arr = np.array(char_image)
            
        char_cache[char] = char_arr
        return char_arr

# TODO: better comparison algorithm
def compare_character(char, block_arr, save_chars, engine):
    height, width = block_arr.shape
    
    char_arr = get_char(char, width, height, save=save_chars)
    if char_arr is None:
        return 0.0
    
    if char_arr.shape != block_arr.shape:
        char_img = Image.fromarray(char_arr)
        char_img = char_img.resize((width, height), Image.Resampling.LANCZOS)
        char_arr = np.array(char_img)
    
    # compare arrays
    if engine == "brightness":
        score = 255.0
        try:
            avg_char = np.mean(char_arr)
            avg_block = np.mean(block_arr)
            score = abs(avg_char - avg_block)
        except Exception as e:
            logging.error(e)
        finally:
            similarity = 1.0 - (score / 255.0)
    elif engine == "ssim":
        score = 0.0
        try:
            min_side = min(block_arr.shape)
            win_size = min(7, min_side)
            if win_size % 2 == 0:
                win_size -= 1
            if win_size < 3:
                win_size = 3
            score = ssim(block_arr, char_arr, data_range=255, win_size=win_size)
        except Exception as e:
            logging.error(e)
            score = 0.0
        finally:
            if isinstance(score, tuple):
                score = score[0]
            try:
                similarity = max(0.0, float(score))
            except Exception:
                similarity = 0.0
    elif engine == "diff":
        pixel_diff = np.sum(np.abs(block_arr - char_arr))

        max_diff = width * height * 255.0
        if max_diff == 0:
            return 1.0

        dissimilarity = pixel_diff / max_diff

        similarity = 1.0 - dissimilarity
    else:
        logging.critical("Invalid Engine specified.")
        sys.exit(1)
        
    return similarity
    
# TODO: implement early stopping
def get_character(img_arr, characters, engine, save_chars):
    max_similarity = 0.0
    best_match = " "
    for character in characters:
        similarity = compare_character(character, img_arr, engine, save_chars)
        if similarity > max_similarity:
            max_similarity = similarity
            best_match = character
            
    return best_match


# will probably not be used but still kept

import matplotlib.font_manager as fm
from PIL import ImageFont


PREFERRED = [
    "DejaVuSansMono",
    "LiberationMono",
    "CourierNew",
    "Menlo",
    "JetBrainsMono",
    "Consolas",
    "Cascadia"
]

STYLE_RANK = {
    "Regular": 0,
    "Book": 1,
    "Normal": 2,
    "Medium": 3,
    "SemiBold": 4,
    "Bold": 5,
    "ExtraBold": 6,
    "Light": 7,
    "ExtraLight": 8,
    "Thin": 9,
}

def get_fonts(preferred=False):
    fonts = []
    seen_names = set()
    font_list = fm.fontManager.ttflist

    for font in font_list:
        if font.name in seen_names:
            continue
        if preferred:
            if any(name.lower() in font.name.lower() for name in PREFERRED) or any(name.lower() in font.fname.lower() for name in PREFERRED):
                fonts.append(font)
                seen_names.add(font.name)
        else:
            fonts.append(font)
            seen_names.add(font.name)

    return [font.fname for font in fonts]

def get_best_candidate(font_paths):    
    # only monospace fonts that aren't italic
    candidates = [
        p for p in font_paths
        if "mono" in p.lower() and "italic" not in p.lower() and "propo" not in p.lower()
    ]
    
    # rank fonts
    def score(path):
        for style, rank in STYLE_RANK.items():
            if style.lower() in path.lower():
                return rank
        return 99

    best = min(candidates, key=score, default=None)
    return best

# very slow, will prob not use
def is_monospace(path, test_chars="ilMW"):
    font = ImageFont.truetype(path, 16)
    widths = [font.getsize(c)[0] for c in test_chars]
    return all(w == widths[0] for w in widths)

def get_monospace_font():
    # get dejavu sans mono path
    for font in fm.fontManager.ttflist:
        if "dejavusansmono" in font.name.replace(" ", "").lower():
            return font.fname

# results are weird
def get_font_aspect(font_path):
    font = ImageFont.truetype(font_path, 16)
    width, height = font.getsize("0")
    return height / width


import os
import sys
import math
import time
import logging
import concurrent.futures

import click
import numpy as np
from PIL import Image, UnidentifiedImageError

from . import __version__
from .core import charsets, unify


logging.basicConfig(
    format="{asctime} - {levelname} - {message}",
    style="{",
    datefmt="%Y-%m-%d %H:%M",
    level=logging.ERROR
)

def process_block(args):
    block, characters, engine, save_chars = args
    return unify.get_character(block, characters, save_chars, engine)

@click.command("koba")
@click.version_option(__version__)
@click.argument(
    "file",
    type=click.Path(
        exists=True,
        file_okay=True,
        readable=True,
    )
)
@click.option(
    "--char-aspect", default=2, show_default=True,
    help="Set character height/width ratio."
)
@click.option(
    "--logging-level",
    default="ERROR",
    show_default=True,
    help="Set logging level. Options: CRITICAL, ERROR, WARNING, INFO, DEBUG."
)
@click.option("--save-blocks", is_flag=True, help="Saves images of blocks to blocks/")
@click.option("--save-chars", is_flag=True, help="Saves images of chars in the specified range to chars/")
@click.option("--engine", "-e", default="brightness", help="Which engine to use for similarity checking (brightness, ssim, diff)", show_default=True)
@click.option("--font", help="Overwrites default font path.", default=None)
@click.option("--char-range", help="Sets the range of unicode symbols to use (start-end)", default="32-128", show_default=True)
def main(file, char_aspect, logging_level, save_blocks, save_chars, engine, font, char_range):
    # update logging level
    logging.getLogger().setLevel(getattr(logging, logging_level.upper(), logging.ERROR))
    
    start_time = time.time()

    split_range = char_range.split("-")
    start_char = split_range[0]
    end_char = split_range[1]
    if start_char.isnumeric() and end_char.isnumeric():
        start_char = int(start_char)
        end_char = int(end_char)
    else:
        logging.critical("Wrong format for char-range.")
        sys.exit(1)
    
    # loading file and reading basic info
    try:
        img = Image.open(file).convert("L")
    except UnidentifiedImageError:
        logging.critical(f"Unsupported or unreadable image format for file: {file}.")
        sys.exit(1)
    img_arr = np.array(img)
    height, width = img_arr.shape[:2]
    logging.debug(f"Image is {width}x{height} pixels.")
    
    # calculating important metrics
    chars_width = os.get_terminal_size().columns
    chars_height = math.ceil((height * chars_width / width) / char_aspect)
    # ppc_w = math.ceil(width / chars_width)
    # ppc_h = math.ceil(height / chars_height)
    
    # distribute extra pixels
    block_widths = [width // chars_width] * chars_width
    for i in range(width % chars_width):
        block_widths[i] += 1

    block_heights = [height // chars_height] * chars_height
    for i in range(height % chars_height):
        block_heights[i] += 1

    logging.debug(f"Image will be {chars_width}x{chars_height} chars.")
    logging.debug(f"Block widths: {block_widths[:5]}... (total {len(block_widths)})")
    logging.debug(f"Block heights: {block_heights[:5]}... (total {len(block_heights)})")
    
    # splitting into blocks with variable sizes
    blocks = []
    y = 0
    for bh in block_heights:
        x = 0
        for bw in block_widths:
            block = img_arr[y:y+bh, x:x+bw]
            blocks.append(block)
            x += bw
        y += bh

    logging.debug(f"{len(blocks)} blocks created.")
    
    if save_blocks:
        logging.info("Saving image blocks...")
        blocks_dir = "blocks"
        os.makedirs(blocks_dir, exist_ok=True)
        for idx, block in enumerate(blocks):
            img_block = Image.fromarray(block)
            img_block.save(os.path.join(blocks_dir, f"{idx:04d}.png"))
    
    characters = charsets.get_range(start_char, end_char)
    all_chars = ""

    # prepare arguments
    block_args = [(block, characters, engine.lower(), save_chars) for block in blocks]

    if font:
        unify.font_path = font

    results = [""] * len(blocks)
    with concurrent.futures.ProcessPoolExecutor() as executor:
        future_to_idx = {
            executor.submit(process_block, arg): idx
            for idx, arg in enumerate(block_args)
        }
        for i, future in enumerate(concurrent.futures.as_completed(future_to_idx), 1):
            idx = future_to_idx[future]
            results[idx] = future.result()
            print(f"Processed {i}/{len(future_to_idx)} blocks", end="\r", flush=True)

    logging.info(f"Took {round(time.time() - start_time, 2)}s")

    all_chars = "".join(results)
    print()
    print(all_chars)



def get_amount(amount):
    chars = []
    i = 32 # first printable ascii
    while len(chars) < amount and i <= 0x10FFFF:
        c = chr(i)
        if c.isprintable() and c.strip():
            chars.append(c)
        i += 1
    return chars

def get_range(start, end):
    chars = []
    for i in range(start, end + 1): # ascii is 32 - 126
        char = chr(i)
        if char.isprintable() and char.strip():
            chars.append(char)
            
    return chars